from .classification import ClassificationTrainer
from .custom_loss import CustomLossTrainer
from .ddpo import DDPOTrainer
from .dpo import DPOTrainer
from .kto import KTOTrainer
from .multigpu import MultiGPUCherryPicksTrainer
from .multimodal import MultimodalTrainer
from .rm import RMTrainer
