from turbo_alignment.modeling.multimodal.projectors.c_abstractor import CAbstractor
from turbo_alignment.modeling.multimodal.projectors.llava import (
    LlavaMultiModalProjector,
)
from turbo_alignment.modeling.multimodal.projectors.registry import (
    MultiModalProjectorRegistry,
)
