from .classification import TrainClassificationStrategy
from .ddpo import TrainDDPOStrategy
from .dpo import TrainDPOStrategy
from .kto import TrainKTOStrategy
from .multimodal import TrainMultimodalStrategy
from .rag import TrainRAGStrategy
from .rm import TrainRMStrategy
from .sft import TrainSFTStrategy
